
/* Web Polygraph       http://www.web-polygraph.org/
 * Copyright 2003-2011 The Measurement Factory
 * Licensed under the Apache License, Version 2.0 */

#ifndef POLYGRAPH__XSTD_RING_H
#define POLYGRAPH__XSTD_RING_H

#include "xstd/Array.h"
#include "xstd/Rnd.h"

// FIFO Queue of <Item>s with fixed capacity (i.e., ring)
// note dynamic resize()-ing on enqueue() does work but is disabled

template <class Item>
class Ring: protected Array<Item> {
	public:
		Ring(int aCapacity = 0): Array<Item>(aCapacity), theInOff(0), theOutOff(0) {}

		int capacity() const { return this->theCapacity; }
		int count() const { return this->theInOff - this->theOutOff; }
		int size() const { return Array<Item>::size(); }
		bool empty() const { return this->theInOff <= this->theOutOff; }
		bool full() const { return count() >= this->theCapacity; }
		const Item &top(int off = 0) const { return item((this->theOutOff+off) % this->theCapacity); }

		Item &top(int off = 0) { return item((this->theOutOff+off) % this->theCapacity); }
		void enqueue(Item i) { Assert(!full()); item(this->theInOff++ % this->theCapacity) = i; check(); }
		Item dequeue() { return item(this->theOutOff++ % this->theCapacity); }

		void reset() { this->theInOff = this->theOutOff = 0; }
		inline void resize(int aCap);

		inline void randomize(RndGen &rng);

		inline Ring &operator =(const Ring &r);

	protected:
		inline void check();
		inline void copyFrom(const Item *const fromItems, const int inOff, const int outOff, const int fromCapacity);

	protected:
		int theInOff;
		int theOutOff;
};


/* implementaion of in-lined methods */

template <class Item>
inline
void Ring<Item>::randomize(RndGen &rng) {
	for (int i = this->theOutOff; i < this->theInOff; ++i)
		Array<Item>::swap(i % this->theCapacity,
			rng(this->theOutOff, this->theInOff) % this->theCapacity);
}

template <class Item>
inline
Ring<Item> &Ring<Item>::operator =(const Ring<Item> &r) {
	if (this->items() != r.items()) {
		reset();
		if (r.capacity() > capacity())
			resize(r.capacity());
		copyFrom(r.items(), r.theInOff, r.theOutOff, r.capacity());
	}
	return *this;
}

template <class Item>
inline
void Ring<Item>::resize(int aCap) {
	const int oldCap = this->theCapacity;
	Item *oldItems = this->theItems;
	this->theItems = 0;

	this->grow(aCap);

	if (oldItems) {
		copyFrom(oldItems, theInOff, theOutOff, oldCap);
		delete[] oldItems;
	}
}

template <class Item>
inline
void Ring<Item>::check() {
	Assert(this->theInOff > this->theOutOff && count() <= this->theCapacity);
	if (this->theOutOff >= this->theCapacity) {
		this->theInOff %= this->theCapacity;
		this->theOutOff %= this->theCapacity;
		if (this->theInOff <= this->theOutOff)
			this->theInOff += this->theCapacity;
	}
}

template <class Item>
inline
void Ring<Item>::copyFrom(const Item *const fromItems, const int inOff, const int outOff, const int fromCapacity) {
	const int count = inOff - outOff;
	Must(outOff <= inOff);
	Must(count <= fromCapacity);
	Must(this->capacity() >= fromCapacity);

	if (count > 0) {
		const int cappedInOff = inOff % fromCapacity;
		const int cappedOutOff = outOff % fromCapacity;
		if (cappedInOff > cappedOutOff)
			this->memmove(0, fromItems + cappedOutOff, count);
		else {
			this->memmove(0, fromItems + cappedOutOff,
				fromCapacity - cappedOutOff);
			this->memmove(fromCapacity - cappedOutOff, fromItems,
				cappedInOff);
		}
	}

	theOutOff = 0;
	theInOff = count;
}

#endif
